#include <vector>
#include <cstdio>

#include "atoms.hh"
#include "level.hh"
#include "pairs.hh"
#include "handling.hh"
#include "rom.hh"

static void LoadScreen(Screen& target,
                       unsigned roomid,
                       int cont_dir)
{
    const Room& room = Rooms[roomid];
    target.cont_dir = cont_dir;
    
    for(unsigned y=0; y<16; ++y)
        for(unsigned x=0; x<16; ++x)
        {
            unsigned blockid = room.blocks[(x/2)*8 + (y/2)];
            const Block& b = Blocks[blockid];
            
            target.outmap[y*16+x] = b.c[(x%2)*2 + (y%2)];
        }
}
                       
static unsigned LoadLevel(Level& target, unsigned order_addr, unsigned layout_addr)
{
    unsigned screen = 0;
    unsigned pos_flags = layout_addr + 1;
    unsigned pos_rooms = order_addr;
    
    for(unsigned a=0; a<MaxRooms; ++a)
        target.rooms[a].cont_dir = Last;

    for(;;)
    {
        unsigned char b = ROM[pos_flags++];
        
        unsigned n_screens = (b & 0x1F);
        for(unsigned a=0; a<=n_screens; ++a, ++screen)
        {
        redo:
            unsigned char c = ROM[pos_rooms++];
            printf("// Room %02X as %u [%02X/%u @ %04X,%04X]\n",
                c, screen, b, a, pos_flags-1, pos_rooms-1);
            if(c >= Rooms.size())
            {
                printf("c=%02X, ignoring\n", c);
                goto redo;
            }
            
            int dir = Right;
            
            if(a == n_screens)
            {
                if(b & 0x20) dir |= Shutter;
                if(b & 0x40) dir |= Down;
                if(b & 0x80) dir |= Up;
            }
            
            if(screen > MaxRooms)
            {
                printf("Too many rooms (%u > %u)\n",
                    screen, MaxRooms);
                continue;
            }

            LoadScreen(target.rooms[screen], c, dir);
            target.CreateAscii(target.rooms[screen]);
        }
        if(!b) break;
    }
    
    return pos_flags - layout_addr;
}

static void DumpLevel(const Level& l)
{
    unsigned levnum = l.LevelNumber;
    printf("static Level %s =\n{\n", l.Name);
    printf("\t%u, // level number\n", levnum);
    printf("\t\"%s\",\n", l.Name);
    printf("\n\t// screens:\n\t{\n");
    for(unsigned a=0; a<MaxRooms; ++a)
    {
        if(l.rooms[a].cont_dir == Last) break;
        
        printf("\t\t{\n");
        for(unsigned y=0; y<16; ++y)
        {
            printf("\t\t\"");
            for(unsigned x=0; x<32; ++x)
            {
                unsigned char asc = l.rooms[a].ascii[y*32+x];
                putchar(asc);
                if(x==15) printf("\"   \"");
            }
            printf("\"");

            /*
            printf("   //");
            for(unsigned x=0; x<16; ++x)
            {
                const Char& chr  = l.rooms[a].outmap[y*16+x];
                printf("%u%02X-", chr.palette, chr.tile);
            }
            */
            printf("\n");
        }
        printf("\t\t,");
        const char* sep = "";
        
        if(l.rooms[a].cont_dir & Up) { printf("%sUp",sep); sep="| "; }
        if(l.rooms[a].cont_dir & Down) { printf("%sDown", sep); sep="| "; }
        if(l.rooms[a].cont_dir & Shutter) { printf("%sShutter", sep); sep="| "; }
        if(!*sep)
        {
            bool last = (a+1)>=MaxRooms || l.rooms[a+1].cont_dir == Last;
            printf(last ? "End": "Right");
        }
        printf("},\n");
    }
    printf("\t},\n\n");
    printf("\t// palette\n\t{\n");
    for(unsigned a=0; a<16*3; ++a)
    {
        if((a&3)==0) printf("\t\t");
        printf("%2u, ", l.Palette[a]);
        if((a&3)==3) printf("\n");
        if((a&15)==15 && a+1<16*3) printf("\n"); //group by 16
    }
    printf("\t},\n\n");
    printf("\t// actives\n\t{\n");
    
    for(unsigned a=0; a<MaxActives; ++a)
    {
        if(l.actives[a].screen == 255) continue;
        printf("\t\tActive( %u, %u, 0x%02X,0x%02X, 0x%02X,0x%02X ),\n",
            l.actives[a].type,
            l.actives[a].screen,
            l.actives[a].x1,
            l.actives[a].y1,
            l.actives[a].x2,
            l.actives[a].y2);
    }
    printf("\t},\n\n");
    printf("\t// enemies\n\t{\n");

    for(unsigned a=0; a<MaxEnemies; ++a)
    {
        if(l.enemies[a].screen == 255) continue;
        printf("\t\tEnemy( %u, 0x%02X,0x%02X, 0x%X ),\n",
            l.enemies[a].screen,
            l.enemies[a].x,
            l.enemies[a].y,
            l.enemies[a].type);
    }
    printf("\t},\n\n");
    
    printf("\t%sAscii, %sChar\n", l.Name, l.Name);
    printf("};\n\n\n");
}

static void LoadPair(LevelPairData& pair)
{
    BaseAddr = pair.BeginAddr;

    Rooms.Load();
    Blocks.Load();
    unsigned lay = BaseAddr + Addr_Layout;
    
    unsigned offset = ROM[Addr_OrderOffsetTable + pair.level1.LevelNumber];
    
    lay += LoadLevel(pair.level1, BaseAddr + Addr_Order + offset, lay)-1;
    pair.level1.ReadPalette(BaseAddr + Addr_Palette1);
    pair.level1.ReadActives(offset, BaseAddr + Addr_Actives1);
    pair.level1.ReadEnemies(offset, Addr_EnemyPointers + pair.level1.LevelNumber*2);

    pair.level1.AnalyzeObjects();
    
    DumpLevel(pair.level1);
    
    if(&pair.level1 != &pair.level2)
    {
        offset = ROM[Addr_OrderOffsetTable + pair.level2.LevelNumber];
        
        lay += LoadLevel(pair.level2, BaseAddr + Addr_Order + offset, lay)-1;
        
        pair.level2.ReadPalette(BaseAddr + Addr_Palette2);
        pair.level2.ReadActives(offset, BaseAddr + Addr_Actives2);
        pair.level2.ReadEnemies(offset, Addr_EnemyPointers + pair.level2.LevelNumber*2);

        pair.level2.AnalyzeObjects();
    
        DumpLevel(pair.level2);
    }
}

int main(void)
{
    LoadPair(pair1);
    LoadPair(pair2);
    LoadPair(pair3);
    LoadPair(pair4);
    LoadPair(pair5);
    LoadPair(pair6);

    return 0;
}
